# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

"""
Provides a auth_required decorator to check a request is authenticated,
and an authorised decorator to check a request is authorised
"""
import functools
import inspect
import logging

from tornado import gen
from tornado.concurrent import Future

from .exceptions import HTTPError


def auth_required(validator):
    """Decorate a RequestHandler or method to require that a request is authenticated

    If decorating a coroutine make sure coroutine decorator is first.
    eg.::

        class Handler(tornado.web.RequestHandler):

            @auth_required(validator)
            @coroutine
            def get(self):
                pass

    :param validator: a coroutine that will validate the token and return True/False
    """
    def _auth_decorator(handler):
        if inspect.isclass(handler):
            return _wrap_class(handler, validator)
        return _auth_required(handler, validator)

    return _auth_decorator


def _wrap_class(request_handler, validator):
    """Decorate each HTTP verb method to check if the request is authenticated

    :param request_handler: a tornado.web.RequestHandler instance
    """
    METHODS = ['get', 'post', 'put', 'head', 'options', 'delete', 'patch']
    for name in METHODS:
        method = getattr(request_handler, name)
        setattr(request_handler, name, _auth_required(method, validator))

    return request_handler


def _get_token(request):
    """
    Gets authentication token from request header
    Will raise 401 error if token not found
    :return token: an authorization token.
    """
    token = request.headers.get('Authorization')
    if not token:
        message = 'Token not in Authorization header'
        logging.warning(message)
        raise HTTPError(401, message)
    return token


def _auth_required(method, validator):
    """Decorate a HTTP verb method and check the request is authenticated

    :param method: a tornado.web.RequestHandler method
    :param validator: a token validation coroutine, that should return
    True/False depending if token is or is not valid
    """
    @gen.coroutine
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        token = _get_token(self.request)
        valid_token = yield validator(token)
        if not valid_token:
            message = 'Invalid token: {}'.format(token)
            logging.warning(message)
            raise HTTPError(401, message)
        result = method(self, *args, **kwargs)

        if isinstance(result, Future):
            result = yield result

        raise gen.Return(result)

    return wrapper


def authorized(validator):
    """Decorate a RequestHandler or method to require that a request is authorized

    If decorating a coroutine make sure coroutine decorator is first.
    eg.::

        class Handler(tornado.web.RequestHandler):

            @authorized(validator)
            @coroutine
            def get(self):
                pass

    :param validator: a coroutine that will authorize the user associated with the token and return True/False
    """
    def _authorized_decorator(method):
        @gen.coroutine
        def wrapper(self, *args, **kwargs):
            token = _get_token(self.request)
            authorized = yield validator(token, **kwargs)
            if not authorized:
                message = 'Token is not authorised for this action: {}'.format(token)
                logging.warning(message)
                raise HTTPError(403, message)

            result = method(self, *args, **kwargs)

            if isinstance(result, Future):
                result = yield result

            raise gen.Return(result)

        return wrapper
    return _authorized_decorator



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

from tornado.options import options
from tornado.web import HTTPError as _HTTPError


class HTTPError(_HTTPError):
    """Subclass of tornado.web.HTTPError.

    Raise a HTTPError to respond to a request with an error. The
    base.BaseHandler will use the information in the exception to form a
    JSON response in our standard error format.

    :param status_code: The status code that should be used in the HTTP response
    :param errors: Error messages
    :param kwargs: optionally include a source of the error (defaults to the
        service's name)
    """
    def __init__(self, status_code, errors, **kwargs):

        super(HTTPError, self).__init__(status_code)
        self.source = kwargs.get('source', getattr(options, 'name', None))
        self.errors = errors



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
#

import logging
import os
import socket
import sys

import click
from chub.api import API
from tornado import httpclient
from tornado.options import options

from . import configure, keygen


@click.group()
def defaults():
    pass


@defaults.command(help='Load fixture data')
def load_data():
    logging.warn('load_data is not implemented')


@defaults.command()
@click.argument('email')
@click.argument('password')
@click.argument('organisation_id')
@click.option('--name', help='The service\'s name')
@click.option('--service_type', help='The service type')
@click.option('--accounts_url', help='URL for the accounts service')
@click.option('--location', help='The url including protocol and port (if required) of service')
@click.option('--config', help='The configuration directory')
def register_service(email, password, organisation_id, name=None,
                     service_type=None, accounts_url=None,
                     location=None, config=None):
    """Register a service with the accounts service

    \b
    EMAIL: a user's email
    PASSWORD: a user's password
    ORGANISATION_ID: ID of the service's parent organisation
    """
    accounts_url = accounts_url or getattr(options, 'url_accounts', None)
    name = name or getattr(options, 'name', None)
    service_type = service_type or getattr(options, 'service_type', None)
    location = location or ('https://localhost:' + str(getattr(options, 'port')))
    config = config or 'config'

    if not accounts_url:
        raise click.ClickException(click.style('accounts_url not defined',
                                               fg='red'))

    if not name:
        raise click.ClickException(click.style('service name not defined',
                                               fg='red'))

    if not service_type:
        raise click.ClickException(click.style('service type not defined',
                                               fg='red'))

    client = API(accounts_url, async=False, validate_cert=False)
    try:
        response = client.accounts.login.post(email=email, password=password)
        client.default_headers['Authorization'] = response.data['token']

        try:
            response = client.accounts.organisations[organisation_id].services.post(
                name=name, location=location, service_type=service_type)
            service_id = response['data']['id']
        except httpclient.HTTPError:
            response = client.accounts.services.get(organisation_id=organisation_id)
            services = [s for s in response['data'] if s['name'] == name]

            if services:
                service_id = services[0]['id']
            else:
                msg = ('Organisation {} does not '
                       'have a service named {}').format(organisation_id, name)
                raise click.ClickException(click.style(msg, fg='red'))

        response = client.accounts.services[service_id].secrets.get()
        client_secrets = response['data']
    except httpclient.HTTPError as exc:
        try:
            msg = exc.response.body
        except AttributeError:
            msg = exc.message
        raise click.ClickException(click.style(msg, fg='red'))
    except socket.error as exc:
        raise click.ClickException(click.style(exc.strerror, fg='red'))

    try:
        with open(os.path.join(config, 'local.conf'), 'r') as f:
            lines = [line for line in f.readlines()
                     if not (line.startswith('service_id') or line.startswith('client_secret'))]
    except IOError:
        lines = []

    lines.append('\nservice_id = "{}"\n'.format(service_id))
    if client_secrets:
        lines.append('client_secret = "{}"\n'.format(client_secrets[0]))
    with open(os.path.join(config, 'local.conf'), 'w') as f:
        f.writelines(lines)

    click.echo(click.style('\n{} service registered\n'.format(name),
                           fg='green'))


def _options():
    """Collect all command line options"""
    opts = sys.argv[1:]
    return [click.Option((v.split('=')[0],)) for v in opts
            if v[0] == '-' and v != '--help']


def run(func):
    """Execute the provided function if there are no subcommands"""

    @defaults.command(help='Run the service')
    @click.pass_context
    def runserver(ctx, *args, **kwargs):
        if (ctx.parent.invoked_subcommand and
                ctx.command.name != ctx.parent.invoked_subcommand):
            return

        # work around the fact that tornado's parse_command_line can't
        # cope with having subcommands / positional arguments.
        sys.argv = [sys.argv[0]] + [a for a in sys.argv if a[0] == '-']

        sys.exit(func())

    return runserver


def _is_command_file(filename):
    return filename.endswith('.py') and filename[:-3] != '__init__'


class Command(click.MultiCommand):
    """A MultiCommand that extends the defaults click.Group with commands
    in a service.

    :param main: the main function to run the service. This function will
                 be called if the CLI is not provided with any subcommands
                 (e.g. `python accounts`).

    :param conf_dir: path to the service's tornado configuration directory

    :param commands_dir: path to the commands directory. Python modules within
                         the module will be attached to this CLI as a subcommand.
                         Each module needs to have a `cli` variable in it's
                         namespace which is a click.Command, click.Group or
                         click.MultiCommand instance.
    """

    def __init__(self, main, conf_dir=None, commands_dir=None, **kwargs):
        self._commands_dir = commands_dir
        if conf_dir:
            configure.load_config_file(conf_dir)

        # This is a bit of a hack, but need to register all parameters from
        # the command line because want to allow tornado to handle them and
        # click doesn't contain an equivalent of the `allow_extra_args`
        # keyword argument that works for options.
        # TODO: don't use tornado command line parsing
        params = _options()

        super(Command, self).__init__(self,
                                      params=params,
                                      callback=run(main),
                                      invoke_without_command=True,
                                      **kwargs)

    def list_commands(self, ctx):
        """List commands from the commands dir and default group"""
        rv = defaults.list_commands(ctx)
        if self._commands_dir:
            for filename in os.listdir(self._commands_dir):
                if _is_command_file(filename) and filename[:-3] not in rv:
                    rv.append(filename[:-3])
            rv.sort()
        return rv

    def get_command(self, ctx, name):
        """Get the command from either the commands dir or default group"""
        if not self._commands_dir:
            return defaults.get_command(ctx, name)

        ns = {}
        fn = os.path.join(self._commands_dir, name + '.py')

        try:
            with open(fn) as f:
                code = compile(f.read(), fn, 'exec')
        except IOError:
            return defaults.get_command(ctx, name)

        eval(code, ns, ns)

        CLI = 'cli'

        try:
            return ns[CLI]
        except KeyError:
            ctx.fail('`{}` not defined in "{}"'.format(CLI, name))


@defaults.command()
@click.argument('name')
@click.option('--dest', help='The directory to put the cert and key in', default='certs')
def create_dev_cert(name, dest):
    """
    A command to generate a self signed certificate for dev purposes.
    :param name: name ro be given to the cert and key
    :param dest: location on local filesystem to store the generated files
    :return:
    """
    if not os.path.exists(dest):
        os.makedirs(dest)
    keygen.gen_ca_cert(name, dest, 3650)


def cli(main, conf_dir=None, commands_dir=None):
    """Convenience function for initialising a Command CLI

    For parameter definitions see :class:`.Command`
    """
    return Command(main, conf_dir=conf_dir, commands_dir=commands_dir)()



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

import json
import logging
import urllib

import tornado.httpclient
from tornado.gen import coroutine, Return
from tornado.web import RequestHandler
from tornado.options import options

from chub import API

from .exceptions import HTTPError


def _get_http_error(kwargs):
    if 'exc_info' in kwargs and hasattr(kwargs['exc_info'][1], 'errors'):
        return kwargs['exc_info'][1]


class CorsHandler(RequestHandler):
    """Shared code for handling CORS"""
    def set_default_headers(self):
        """Set the default headers

        If the 'cors' option is True, will set the 'Access-Control-Allow-Origin'
        header to '*'
        """
        if getattr(options, 'cors', False):
            # Allow requests from any origin, can implement white-list later
            self.set_header('Access-Control-Allow-Origin', '*')

    def options(self, *args, **kwargs):
        """Default OPTIONS response

        If the 'cors' option is True, will respond with an empty response and
        set the 'Access-Control-Allow-Headers' and
        'Access-Control-Allow-Methods' headers
        """
        if getattr(options, 'cors', False):
            self.set_header('Access-Control-Allow-Headers',
                            'Content-Type, Authorization, '
                            'Accept, X-Requested-With')
            self.set_header('Access-Control-Allow-Methods',
                            'OPTIONS, TRACE, GET, HEAD, POST, '
                            'PUT, PATCH, DELETE')

        self.finish()


class JsonHandler(RequestHandler):
    """Shared code for handling JSON requests"""
    def write_error(self, status_code, **kwargs):
        """Override `write_error` in order to output JSON errors

        :param status_code: the response's status code, e.g. 500
        """
        http_error = _get_http_error(kwargs)
        if http_error:
            self.finish(self._error_template(status_code,
                                             http_error.errors,
                                             http_error.source))
        else:
            source = kwargs.get('source', getattr(options, 'name', None))
            # Slightly annoyed that have to rely on the internal self._reason
            # to deal with unhandled exceptions. On the dev version of
            # tornado self._reason is always set, while in the current version
            # a reason kwarg is passed down from `send_error` but not set
            # on the instance.
            reason = kwargs.get('reason', self._reason)
            self.finish(self._error_template(status_code, reason, source))

    @classmethod
    def _error_template(cls, status_code, errors, source=None):
        """Construct JSON error response

        :param status_code: the http status code
        :param errors: string or list of error strings
        :param source: source of the error
        :returns: dictionary, e.g.
            {
                'status': 400,
                'errors': [
                    {
                        'source': 'accounts' ,
                        'message':'errormsg1'
                    },
                    {
                        'source': 'accounts',
                        'message':'errormsg2'
                    }
                ]
            }
        """
        # this handles unhandled exceptions
        if isinstance(errors, basestring):
            errors_out = {'errors': [{'message': errors}]}
        elif isinstance(errors, (list, tuple)):
            errors_out = {'errors': [{'message': e} for e in errors]}
        else:
            errors_out = errors
        errors_out['status'] = status_code

        for error in errors_out['errors']:
            if not error.get('source'):
                error['source'] = source

        logging.error(json.dumps(errors_out))
        return errors_out

    def get_json_body(self, required=None, validators=None):
        """Get JSON from the request body

        :param required: optionally provide a list of keys that should be
        in the JSON body (raises a 400 HTTPError if any are missing)
        :param validator: optionally provide a dictionary of items that should
        be in the body with a method that validates the item.
        The method must be synchronous and return a boolean, no exceptions.
        :raises: HTTPError
        """
        content_type = self.request.headers.get('Content-Type',
                                                'application/json')
        if 'application/json' not in content_type.split(';'):
            raise HTTPError(415, 'Content-Type should be application/json')
        if not self.request.body:
            error = 'Request body is empty'
            logging.warning(error)
            raise HTTPError(400, error)

        try:
            body = json.loads(self.request.body)
        except (ValueError, TypeError):
            error = 'Error parsing JSON'
            logging.warning(error)
            raise HTTPError(400, error)

        if required:
            _check_required(body, required)

        if validators:
            _validate(body, validators)

        return body


class AuthHandler(RequestHandler):
    UNAUTHENTICATED_ACCESS = "unauthenticated"
    READ_ACCESS = "r"
    WRITE_ACCESS = "w"
    READ_WRITE_ACCESS = "rw"

    # Default required access levels for handler methods.
    # Can be overridden by individual service handlers
    METHOD_ACCESS = {
        "GET": READ_ACCESS,
        "HEAD": READ_ACCESS,
        "POST": WRITE_ACCESS,
        "PATCH": WRITE_ACCESS,
        "PUT": WRITE_ACCESS,
        "DELETE": WRITE_ACCESS
    }

    def endpoint_access(self, method):
        """
        Determine access level needed for endpoint
        :param method: The request verb
        :return: String representing access type.
        """
        if method == 'OPTIONS':
            # The CORS pre-flight checks should not require authentication
            return self.UNAUTHENTICATED_ACCESS
        elif method not in self.METHOD_ACCESS:
            logging.error('Cannot determine access needed for %s method', method)
            raise HTTPError(500, 'Internal Server Error')

        return self.METHOD_ACCESS[method]

    @coroutine
    def verify_token(self, token, requested_access):
        """
        Check the token bearer is permitted to access the resource

        :param token: Access token
        :param requested_access: the access level the client has requested
        :returns: boolean
        """
        client = API(options.url_auth,
                     auth_username=options.service_id,
                     auth_password=options.client_secret,
                     ca_certs=options.ssl_ca_cert)
        headers = {'Content-Type': 'application/x-www-form-urlencoded',
                   'Accept': 'application/json'}
        body = urllib.urlencode({'token': token, 'requested_access': requested_access})

        client.auth.verify.prepare_request(headers=headers, request_timeout=180)

        try:
            result = yield client.auth.verify.post(body=body)
        except tornado.httpclient.HTTPError as ex:
            # Must be converted to a tornado.web.HTTPError for the server
            # to handle it correctly
            logging.exception(ex.message)
            raise HTTPError(500, 'Internal Server Error')

        raise Return(result['has_access'])

    @coroutine
    def prepare(self):
        """If OAuth verification is required, validate provided token

        :raise: HTTPError if token does not have access
        """
        requested_access = self.endpoint_access(self.request.method)
        use_oauth = getattr(options, 'use_oauth', None)
        if use_oauth and requested_access is not self.UNAUTHENTICATED_ACCESS:
            token = self.request.headers.get('Authorization', '').split(' ')[-1]
            if token:
                has_access = yield self.verify_token(token, requested_access)
                if not has_access:
                    msg = "'{}' access not granted.".format(requested_access)
                    raise HTTPError(403, msg)
            else:
                msg = 'OAuth token not provided'
                raise HTTPError(401, msg)


class BaseHandler(AuthHandler, CorsHandler, JsonHandler):
    pass


def _check_required(body, keys):
    msg = 'Please provide a {}'
    errors = [msg.format(k) for k in keys if k not in body]

    if errors:
        raise HTTPError(400, errors)


def _validate(body, validators):
    msg = 'Invalid data element {}'
    errors = [msg.format(key) for key, func in validators.items()
              if not func(body.get(key))]

    if errors:
        raise HTTPError(400, errors)



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

"""Useful utils."""
from functools import wraps
import string

from tornado.web import RedirectHandler
from tornado.options import options
from koi.constants import PATH_PART


def tuplify(*args):
    """
    Convert args to a tuple, unless there's one arg and it's a
    function, then acts a decorator.
    """
    if (len(args) == 1) and callable(args[0]):
        func = args[0]

        @wraps(func)
        def _inner(*args, **kwargs):
            return tuple(func(*args, **kwargs))
        return _inner
    else:
        return tuple(args)


def listify(*args):
    """
    Convert args to a list, unless there's one arg and it's a
    function, then acts a decorator.
    """
    if (len(args) == 1) and callable(args[0]):
        func = args[0]

        @wraps(func)
        def _inner(*args, **kwargs):
            return list(func(*args, **kwargs))
        return _inner
    else:
        return list(args)


def stringify(*args):
    """
    Joins args to build a string, unless there's one arg and it's a
    function, then acts a decorator.
    """
    if (len(args) == 1) and callable(args[0]):
        func = args[0]

        @wraps(func)
        def _inner(*args, **kwargs):
            return "".join([str(i) for i in func(*args, **kwargs)])
        return _inner
    else:
        return "".join([str(i) for i in args])


def add_prefix(endpoints, prefix, kwargs=None):
    for endpoint in endpoints:
        url = endpoint[0]
        url = add_path_part(url)
        if kwargs:
            k = kwargs.copy()
            if len(endpoint) >= 3:
                # assume that the kwargs in the endpoint should over-ride
                # the kwargs passed in as an argument
                k.update(endpoint[2])

            yield tuplify(prefix+url, endpoint[1], k, *endpoint[3:])
        else:
            yield tuplify(prefix+url, *endpoint[1:])


@stringify
def add_path_part(url, regex=PATH_PART):
    """
    replace the variables in a url template with regex named groups
    :param url: string of a url template
    :param regex: regex of the named group
    :returns: regex
    """
    formatter = string.Formatter()
    url_var_template = "(?P<{var_name}>{regex})"

    for part in formatter.parse(url):
        string_part, var_name, _, _ = part
        if string_part:
            yield string_part
        if var_name:
            yield url_var_template.format(var_name=var_name, regex=regex)


@listify
def make_endpoints(version, name, endpoints, kwargs=None):
    """
    Returns a redirect handler and all endpoints with a version prefix added.

    :param version: the application version
    :param name: the application name
    :param endpoints: a list of application endpoints
    :param kwargs: an optional dictionary to populate placeholders in endpoints
    :returns:list of endpoints
    """
    if kwargs is None:
        kwargs = {}

    version_url_prefix = '/v{}/{}'.format(version.split('.')[0], name)
    yield (r"/", RedirectHandler, {"url": r"{}".format(version_url_prefix)})
    for endpoint in add_prefix(endpoints, version_url_prefix, kwargs):
        yield endpoint


REQUIRED_CAPABILITIES = ['service_id']
PROTECTED_CAPABILITIES = ['client_secret']


def sanitise_capabilities(capabilities):
    """
    Makes sure dictionary of capabilities includes required options, and does not include protected ones.
    :param capabilities:
    :return: dict
    """
    for c in REQUIRED_CAPABILITIES:
        capabilities[c] = options[c]

    for c in PROTECTED_CAPABILITIES:
        if c in capabilities:
            del capabilities['client_secret']

    return capabilities



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 


import functools

from tornado.ioloop import IOLoop
from tornado.gen import coroutine
from tornado.concurrent import Future


def gen_test(func):
    """
    Helper for running async tests, based on the tornado.testing.gen_test
    decorator.
    It wraps the test function with tornado.gen.coroutine and initialises an
    IOLoop to run the async code using IOLoop.run_sync
    NOTE: if using this with the mock.patch decorator apply gen_test first,
    otherwise the patches won't work.
    ANOTHER NOTE: if you don't yield when calling coroutines in your test you
    can get false positives, just like any other place where you don't call
    coroutines correctly. It's always a good idea to see your test fail so
    you know it's testing something.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        cofunc = coroutine(func)
        io_loop = IOLoop.current()

        try:
            result = io_loop.run_sync(functools.partial(cofunc, *args, **kwargs))
            return result

        finally:
            io_loop.clear_current()
            if not IOLoop.initialized() or io_loop is not IOLoop.instance():
                io_loop.close(all_fds=True)


    return wrapper

def make_future(result):
    """Create a `tornado.concurrent.Future` that returns `result`

    Useful for adding a return value to a mocked coroutine, for example::

        mock = Mock()
        mock.func.return_value = test_helpers.make_future('test')
        result = IOLoop.instance().run_sync(mock.func)

        assert result == 'test'

    :param result: the Future's result
    :returns: tornado.concurrent.Future
    """
    f = Future()
    f.set_result(result)

    return f



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

from .configure import (
    define_options, load_config, make_server, make_application)
from .keygen import (
    CA_CRT, CA_KEY, LOCALHOST_CRT, LOCALHOST_KEY, CLIENT_CRT, CLIENT_KEY,
    MOZILLA_PEM)
from .auth import auth_required
from . import test_helpers
from . import base

__version__ = '1.0.8'



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

import os
import subprocess
import argparse
import tempfile

DEFAULT_CERTS_DIR = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '../certs'))

CA_CRT = os.path.join(DEFAULT_CERTS_DIR, 'CA.crt')
CA_KEY = os.path.join(DEFAULT_CERTS_DIR, 'CA.key')
MOZILLA_PEM = os.path.join(DEFAULT_CERTS_DIR, 'ca-bundle.crt')
LOCALHOST_CRT = os.path.join(DEFAULT_CERTS_DIR, 'localhost.crt')
LOCALHOST_KEY = os.path.join(DEFAULT_CERTS_DIR, 'localhost.key')
CLIENT_CRT = os.path.join(DEFAULT_CERTS_DIR, 'client.crt')
CLIENT_KEY = os.path.join(DEFAULT_CERTS_DIR, 'client.key')

SUBJECT = ('/C=GB/ST=London/L=London'
           '/O=Connected Digital Economy Catapult'
           '/OU=Dev Team'
           '/CN=Open Permissions Platform Coalition')

SUBJECT_ALT_NAME = '''\
[req]
req_extensions = v3_req
distinguished_name = req_distinguished_name

[v3_req]
basicConstraints = CA:FALSE
keyUsage = nonRepudiation, digitalSignature, keyEncipherment
subjectAltName = @alt_names

[req_distinguished_name]
# we get this from the command line through -subj

[alt_names]
'''


def call_openssl(cmd, message, silent=False):
    """
    call openssl
    :param cmd: a string of command send to openssl
    :param message: a string to print out if not silent
    :param silent: a boolean for whether to suppress output from openssl
    """
    if silent:
        with open(os.devnull, 'w') as devnull:
            return subprocess.check_call(cmd, shell=True, stdout=devnull,
                                         stderr=subprocess.STDOUT)
    else:
        print message
        return subprocess.check_call(cmd, shell=True)


def gen_private_key(filepath, silent=False):
    """
    generate ssl private key
    :param filepath: file path to the key file
    :param silent: whether to suppress output
    """
    cmd = 'openssl genrsa -out {} 2048'.format(filepath)
    message = 'generate private key {}'.format(filepath)
    call_openssl(cmd, message, silent)


def gen_self_signed_cert(filepath, keyfile, days, silent=False):
    """
    generate self signed ssl certificate, i.e. a private CA certificate
    :param filepath: file path to the key file
    :param keyfile: file path to the private key
    :param days: valid duration for the certificate
    :param silent: whether to suppress output
    """
    cmd = (
        'openssl req -x509 -new -nodes -key {} -days {} -out {} -subj "{}"'
    ).format(keyfile, days, filepath, SUBJECT)
    message = 'generate self signed certificate {} for {} days'.format(
        filepath, days)
    call_openssl(cmd, message, silent)


def gen_cert_request(filepath, keyfile, config, silent=False):
    """
    generate certificate request
    :param filepath: file path to the certificate request
    :param keyfile: file path to the private key
    :param silent: whether to suppress output
    """
    message = 'generate ssl certificate request'
    cmd = (
        'openssl req -new -key {} -out {} -subj "{}"'
        ' -extensions v3_req -config {}').format(
            keyfile, filepath, SUBJECT, config)
    call_openssl(cmd, message, silent)


def sign_cert_request(filepath, cert_req, ca_crt, ca_key, days, extfile,
                      silent=False):
    """
    generate self signed ssl certificate, i.e. a private CA certificate
    :param filepath: file path to the key file
    :param keyfile: file path to the private key
    :param days: valid duration for the certificate
    :param silent: whether to suppress output
    """
    message = 'sign certificate request'
    cmd = ('openssl x509 -req -in {} -CA {} -CAkey {} -CAcreateserial'
           ' -out {} -days {} -extfile {} -extensions v3_req').format(
               cert_req, ca_crt, ca_key, filepath, days, extfile)
    call_openssl(cmd, message, silent)


def gen_ca_cert(filename, dirname, days, silent=False):
    """
    generate a CA key and certificate key pair.
    :param filename: prefix for the key and cert file
    :param dirname: name of the directory
    :param days: days of the certificate being valid
    :param silent: whether to suppress output
    """
    keyfile = os.path.join(dirname, '{}.key'.format(filename))
    ca_crt = os.path.join(dirname, '{}.crt'.format(filename))
    gen_private_key(keyfile, silent)
    gen_self_signed_cert(ca_crt, keyfile, days, silent)


def gen_non_ca_cert(filename, dirname, days, ip_list, dns_list,
                    ca_crt, ca_key, silent=False):
    """
    generate a non CA key and certificate key pair signed by the private
    CA key and crt.
    :param filename: prefix for the key and cert file
    :param dirname: name of the directory
    :param days: days of the certificate being valid
    :ip_list: a list of ip address to be included in the certificate
    :dns_list: a list of dns names to be included in the certificate
    :ca_key: file path to the CA key
    :ca_crt: file path to the CA crt
    :param silent: whether to suppress output
    """
    key_file = os.path.join(dirname, '{}.key'.format(filename))
    req = os.path.join(dirname, '{}.csr'.format(filename))
    crt = os.path.join(dirname, '{}.crt'.format(filename))
    gen_private_key(key_file, silent)
    alt_names = []

    for ind, ip in enumerate(ip_list):
        alt_names.append('IP.{} = {}'.format(ind + 1, ip))
    for ind, dns in enumerate(dns_list):
        alt_names.append('DNS.{} = {}'.format(ind + 1, dns))

    conf = tempfile.mktemp()
    open(conf, 'w').write(SUBJECT_ALT_NAME + '\n'.join(alt_names))
    gen_cert_request(req, key_file, conf, silent)
    sign_cert_request(crt, req, ca_crt, ca_key, days, conf, silent)


def check_cert(certfile):
    """
    output the text format of the certificate
    :param filepath: file path to the ssl certificate
    :returns: string
    """
    cmd = 'openssl x509 -in {} -text -noout'.format(certfile)
    out = subprocess.check_output(cmd, shell=True)
    return out


def check_key_cert_match(keyfile, certfile):
    """
    check if the ssl key matches the certificate
    :param keyfile: file path to the ssl key
    :param certfile: file path to the ssl certificate
    :returns: true or false
    """
    key_modulus = subprocess.check_output(
        'openssl rsa -noout -modulus -in {}'.format(keyfile),
        shell=True)
    cert_modulus = subprocess.check_output(
        'openssl x509 -noout -modulus -in {}'.format(certfile),
        shell=True)
    return key_modulus == cert_modulus


def argument_parser():
    """
    process the command line arguments.
    :returns: an ArgumentParser object.
    """
    parser = argparse.ArgumentParser(
        description='openssl wrapper')
    parser.add_argument('--ca', action='store_true', default=False,
                        dest='is_ca',
                        help='Is it a Certificate Authority certificate?')
    parser.add_argument('-f', '--filename', type=str, default='localhost',
                        help='File name prefix of the key and certificate.')
    parser.add_argument('-d', '--dirname', type=str, default=DEFAULT_CERTS_DIR,
                        help='Destination directory name.')
    parser.add_argument('--days', type=int, default=500,
                        help='Duration of certificate.')
    parser.add_argument('--ip', nargs='*', default=['127.0.0.1'],
                        help='IP address.')
    parser.add_argument('--dns', nargs='*', default=['localhost'],
                        help='DNS')
    parser.add_argument('--silent', action='store_true',
                        help='should the output from the shell be silenced?')
    return parser


def main(argv=None):
    args = argument_parser().parse_args(argv)
    if args.is_ca:
        gen_ca_cert(args.filename, args.dirname, args.days, args.silent)
    else:
        ca_crt = os.path.join(DEFAULT_CERTS_DIR, 'CA.crt')
        ca_key = os.path.join(DEFAULT_CERTS_DIR, 'CA.key')
        gen_non_ca_cert(args.filename, args.dirname, args.days,
                        args.ip, args.dns, ca_crt, ca_key, args.silent)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# Copyright 2016 Open Permissions Platform Coalition
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy of the License at
# http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software distributed under the License is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and limitations under the License.
# 

"""
Configure ssl server and client
"""
from datetime import datetime
import hashlib
import logging
from logging.handlers import SysLogHandler
import os
import socket
import ssl
import time

import sys
import tornado.httpserver
from tornado.options import define, options, parse_command_line
from tornado.util import exec_in
from tornado.escape import native_str
from tornado.web import RequestHandler

from .keygen import DEFAULT_CERTS_DIR
from .utils import make_endpoints

general_logger = logging.getLogger('tornado.general')


def _default_certs_path(name):
    """
    prefix with the path to the certs directory
    """
    return os.path.join(DEFAULT_CERTS_DIR, name)


define('ssl_ca_cert', default=_default_certs_path('CA.crt'))
define('ssl_key', default=_default_certs_path('localhost.key'))
define('ssl_cert', default=_default_certs_path('localhost.crt'))
define('ssl_cert_reqs', default=0)


def define_options(default_conf):
    """
    Define the options from default.conf dynamically
    """
    default = {}
    with open(default_conf, 'rb') as f:
        exec_in(native_str(f.read()), {}, default)

    for name, value in default.iteritems():
        # if the option is already defined by tornado
        # override the value
        # a list of options set by tornado:
        # log_file_num_backups, logging, help,
        # log_to_stderr, log_file_max_size, log_file_prefix
        if name in options:
            setattr(options, name, value)
        # otherwise define the option
        else:
            define(name, value)


def log_config():
    """Logs the config used to start the application"""
    conf = '\n'.join(
            ['{}="{}"'.format(k, v) for k, v
             in sorted(options.as_dict().iteritems())])
    logging.info('Service started with the following settings:\n' + conf)


def ssl_server_options():
    """
    ssl options for tornado https server
    these options are defined in each application's default.conf file
    if left empty, use the self generated keys and certificates included
    in this package.
    this function is backward compatible with python version lower than
    2.7.9 where ssl.SSLContext is not available.
    """
    cafile = options.ssl_ca_cert
    keyfile = options.ssl_key
    certfile = options.ssl_cert
    verify_mode = options.ssl_cert_reqs
    try:
        context = ssl.create_default_context(
            purpose=ssl.Purpose.CLIENT_AUTH, cafile=cafile)
        context.load_cert_chain(certfile=certfile, keyfile=keyfile)
        context.verify_mode = verify_mode
        return context
    except AttributeError:
        ssl_options = {
            'ca_certs': cafile,
            'keyfile': keyfile,
            'certfile': certfile,
            'cert_reqs': verify_mode
        }
        return ssl_options


class RequestFilter(logging.Filter):
    def __init__(self, request):
        self.request = request
        self.request_id = hashlib.md5(datetime.now().isoformat() +
                                      request.remote_ip).hexdigest()[:10]

    def filter(self, record):
        record.remote_ip = self.request.remote_ip
        record.request_id = self.request_id
        return True


def log_formatter(request=None):
    """
    Log formatter used in our syslog

    :param request: a request object
    :returns: logging.Formatter
    """

    if request:
        format_str = ('%(asctime)s {ip} {name}:  ENV={env} '
                      'REMOTE_IP=%(remote_ip)s REQUEST_ID=%(request_id)s '
                      '%(message)s')
    else:
        format_str = '%(asctime)s {ip} {name}:  ENV={env} %(message)s'

    try:
        hostname = socket.gethostname()
    except socket.gaierror:
        hostname = 'localhost'

    try:
        ip = socket.gethostbyname(hostname)
    except socket.gaierror:
        ip = '127.0.0.1'

    formatter = logging.Formatter(
            format_str.format(ip=ip, name=options.name, env=options.env),
            datefmt='%Y-%m-%dT%H:%M:%S')
    logging.Formatter.converter = time.gmtime

    return formatter


def configure_syslog(request=None, logger=None, exceptions=False):
    """
    Configure syslog logging channel.
    It is turned on by setting `syslog_host` in the config file.
    The port default to 514 can be overridden by setting `syslog_port`.

    :param request: tornado.httputil.HTTPServerRequest instance
    :param exceptions: boolean - This indicates if we should raise
        exceptions encountered in the logging system.
    """
    syslog_host = getattr(options, 'syslog_host', None)
    if not syslog_host:
        return

    sys.modules["logging"].raiseExceptions = exceptions
    handler = SysLogHandler(address=(syslog_host, options.syslog_port))
    formatter = log_formatter(request)
    handler.setFormatter(formatter)

    if request:
        handler.addFilter(RequestFilter(request))

    if logger:
        logger.addHandler(handler)
    else:
        logging.getLogger().addHandler(handler)


def load_config_file(conf_dir):
    default_conf = os.path.join(conf_dir, 'default.conf')
    define_options(default_conf)
    local_conf = os.path.join(conf_dir, 'local.conf')
    if os.path.isfile(local_conf):
        options.parse_config_file(local_conf, final=False)


def load_config(conf_dir):
    """
    Use default.conf as the definition of options with default values
    using tornado.options.define.
    Then overrides the values from: local.conf.
    This mapping allows to access the application configuration across the
    application.

    :param conf_dir: path to configuration directory
    """
    load_config_file(conf_dir)
    # NOTE:
    # logging before this line is not going to work
    parse_command_line(final=True)


class ErrorHandler(RequestHandler):
    def prepare(self):
        self.set_status(404)
        self.finish({'status': 404, 'errors': [{'message': 'Not Found'}]})


def make_application(version, app_name, app_urls, kwargs=None):
    """
    Loads the routes and starts the server

    :param version: the application version
    :param app_name: the application name
    :param app_urls: a list of application endpoints
    :param kwargs: dictionary of options
    :returns: tornado.web.Application instance
    """
    if kwargs is None:
        kwargs = {}

    urls = make_endpoints(version, app_name, app_urls, kwargs)
    application = tornado.web.Application(
            urls,
            default_handler_class=kwargs.get('default_handler_class', ErrorHandler))

    return application


def make_server(application, conf_dir=None):
    """
    Configure the server return the server instance
    """
    if conf_dir:
        load_config(conf_dir)
    configure_syslog()
    log_config()
    if options.use_ssl:
        ssl_options = ssl_server_options()
        server = tornado.httpserver.HTTPServer(
                application, ssl_options=ssl_options)
        general_logger.info(
                'start tornado https server at https://%s:%s'
                ' with ssl_options: %s', options.ip, options.port, ssl_options)
    else:

        server = tornado.httpserver.HTTPServer(application)
        general_logger.info('start tornado http server at http://{0}:{1}'.format(
                options.ip, options.port))

    server.bind(options.port, options.ip)
    return server
